namespace PrayerTracker.Data

/// Table names
[<RequireQualifiedAccess>]
module Table =

    /// The church table
    [<Literal>]
    let Church = "church"

    /// The small group table
    [<Literal>]
    let Group = "small_group"

    /// The small group member table
    [<Literal>]
    let Member = "member"

    /// The prayer request table
    [<Literal>]
    let Request = "prayer_request"

    /// The user table
    [<Literal>]
    let User = "pt_user"


open System
open NodaTime
open PrayerTracker.Entities

/// JSON serialization customizations
[<RequireQualifiedAccess>]
module Json =

    open System.Text.Json.Serialization

    /// Convert a wrapped DU to/from its string representation
    type WrappedJsonConverter<'T>(wrap: string -> 'T, unwrap: 'T -> string) =
        inherit JsonConverter<'T>()
        override _.Read(reader, _, _) = wrap (reader.GetString())
        override _.Write(writer, value, _) = writer.WriteStringValue(unwrap value)

    open System.Text.Json
    open NodaTime.Serialization.SystemTextJson

    /// JSON serializer options to support the target domain
    let options =
        let opts = JsonSerializerOptions()

        [ WrappedJsonConverter<AsOfDateDisplay>(AsOfDateDisplay.Parse, string) :> JsonConverter
          WrappedJsonConverter<EmailFormat>(EmailFormat.Parse, string)
          WrappedJsonConverter<Expiration>(Expiration.Parse, string)
          WrappedJsonConverter<PrayerRequestType>(PrayerRequestType.Parse, string)
          WrappedJsonConverter<RequestSort>(RequestSort.Parse, string)
          WrappedJsonConverter<TimeZoneId>(TimeZoneId, string)
          WrappedJsonConverter<ChurchId>(Guid.Parse >> ChurchId, string)
          WrappedJsonConverter<MemberId>(Guid.Parse >> MemberId, string)
          WrappedJsonConverter<PrayerRequestId>(Guid.Parse >> PrayerRequestId, string)
          WrappedJsonConverter<SmallGroupId>(Guid.Parse >> SmallGroupId, string)
          WrappedJsonConverter<UserId>(Guid.Parse >> UserId, string)
          JsonFSharpConverter() ]
        |> List.iter opts.Converters.Add

        let _ = opts.ConfigureForNodaTime DateTimeZoneProviders.Tzdb
        opts.PropertyNamingPolicy <- JsonNamingPolicy.CamelCase
        opts.DefaultIgnoreCondition <- JsonIgnoreCondition.WhenWritingNull
        opts


open BitBadger.Documents
open BitBadger.Documents.Sqlite

/// Establish the required data environment
[<RequireQualifiedAccess>]
module Connection =

    open System.Text.Json

    /// Ensure tables and indexes are defined
    let setUp () =
        backgroundTask {
            Configuration.useIdField "id"

            Configuration.useSerializer
                { new IDocumentSerializer with
                    member _.Serialize<'T>(it: 'T) =
                        JsonSerializer.Serialize(it, Json.options)

                    member _.Deserialize<'T>(it: string) =
                        JsonSerializer.Deserialize<'T>(it, Json.options) }

            let! tables = Custom.list<string> "SELECT name FROM sqlite_master WHERE type = 'table'" [] _.GetString(0)

            if not (List.contains Table.Church tables) then
                do! Definition.ensureTable Table.Church

            if not (List.contains Table.Group tables) then
                do! Definition.ensureTable Table.Group
                do! Definition.ensureFieldIndex Table.Group "church" [ "churchId" ]

            if not (List.contains Table.Member tables) then
                do! Definition.ensureTable Table.Member
                do! Definition.ensureFieldIndex Table.Member "group" [ "smallGroupId" ]

            if not (List.contains Table.Request tables) then
                do! Definition.ensureTable Table.Request
                do! Definition.ensureFieldIndex Table.Request "group" [ "smallGroupId" ]

            if not (List.contains Table.User tables) then
                do! Definition.ensureTable Table.User
                do! Definition.ensureFieldIndex Table.User "email" [ "email" ]
        }


open Microsoft.Data.Sqlite

/// Functions to retrieve small group information
module SmallGroups =

    /// Query to retrieve data for a small group info instance
    let private infoQuery =
        $"SELECT g.data->>'id' AS id, g.data->>'groupName' AS groupName, c.data->>'churchName' AS churchName,
                 g.data->'preferences'->>'timeZoneId' AS timeZoneId, g.data->'preferences'->>'isPublic' AS isPublic
            FROM {Table.Group} g
                 INNER JOIN {Table.Church} c ON c.data->>'id' = g.data->>'churchId'"

    /// Query to retrieve data for a small group select list item
    let private itemQuery =
        $"SELECT g.data->>'groupName' AS groupName, g.data->>'id' AS id, c.data->>'churchName' AS churchName
            FROM {Table.Group} g
                 INNER JOIN {Table.Church} c ON c.data->>'id' = g.data->>'churchId'"

    /// The ORDER BY clause for select list item queries
    let private itemOrderBy = "ORDER BY c.data->>'churchName', g.data->>'groupName'"

    /// Map a row to a Small Group list item
    let private toSmallGroupItem (rdr: SqliteDataReader) =
        (rdr.GetOrdinal >> rdr.GetString >> Guid.Parse >> Giraffe.ShortGuid.fromGuid) "id",
        $"""{(rdr.GetOrdinal >> rdr.GetString) "churchName"} | {(rdr.GetOrdinal >> rdr.GetString) "groupName"}"""

    /// Get the group IDs for the given church
    let internal groupIdsByChurch (churchId: ChurchId) =
        backgroundTask {
            let! groups = Find.byFields<SmallGroup> Table.Group All [ Field.Equal "churchId" churchId ]
            return groups |> List.map _.Id
        }

    /// Count the number of small groups for a church
    let countByChurch (churchId: ChurchId) =
        Count.byFields Table.Group All [ Field.Equal "churchId" churchId ]

    /// Delete a small group by its ID
    let deleteById (groupId: SmallGroupId) =
        backgroundTask {
            use conn = Configuration.dbConn ()
            use! txn = conn.BeginTransactionAsync()

            let! users = Find.byFields<User> Table.User All [ Field.InArray "smallGroups" Table.User [ groupId ] ]

            for user in users do
                do! Patch.byId Table.User user.Id {| SmallGroups = user.SmallGroups |> List.except [ groupId ] |}

            do! conn.deleteByFields Table.Request All [ Field.Equal "smallGroupId" groupId ]
            do! conn.deleteById Table.Group groupId

            do! txn.CommitAsync()
        }

    /// Get information for all small groups
    let infoForAll () =
        Custom.list $"{infoQuery} ORDER BY g.data->>'groupName'" [] SmallGroupInfo.FromReader

    /// Get a list of small group IDs along with a description that includes the church name
    let listAll () =
        Custom.list $"{itemQuery} {itemOrderBy}" [] toSmallGroupItem

    /// Get a list of small group IDs and descriptions for groups with a group password
    let listProtected () =
        Custom.list
            $"{itemQuery} WHERE COALESCE(g.data->'preferences'->>'groupPassword', '') <> '' {itemOrderBy}"
            []
            toSmallGroupItem

    /// Get a list of small group IDs and descriptions for groups that are public or have a group password
    let listPublicAndProtected () =
        Custom.list
            $"{infoQuery}
              WHERE g.data->'preferences'->>'isPublic' = TRUE
                 OR COALESCE(g.data->'preferences'->>'groupPassword', '') <> ''
              ORDER BY c.data->>'churchName', g.data->>'groupName'"
            []
            SmallGroupInfo.FromReader

    /// Log on for a small group (includes list preferences)
    let logOn (groupId: SmallGroupId) (password: string) =
        Find.firstByFields<SmallGroup>
            Table.Group
            All
            [ Field.Equal "id" groupId; Field.Equal "preferences.groupPassword" password ]

    /// Save a small group
    let save group = save<SmallGroup> Table.Group group

    /// Save a small group's list preferences
    let savePreferences (groupId: SmallGroupId) (pref: ListPreferences) =
        Patch.byId Table.Group groupId {| Preferences = pref |}

    /// Get a small group by its ID (including list preferences)
    let tryById groupId =
        Find.byId<SmallGroupId, SmallGroup> Table.Group groupId


/// Functions to manipulate churches
module Churches =

    /// Get a list of all churches
    let all () = Find.all<Church> Table.Church

    /// Delete a church by its ID
    let deleteById churchId =
        backgroundTask {
            use conn = Configuration.dbConn ()
            use! txn = conn.BeginTransactionAsync()

            let! groupIds = SmallGroups.groupIdsByChurch churchId

            do! Delete.byFields Table.Request All [ Field.In "smallGroupId" groupIds ]

            let! users = Find.byFields<User> Table.User All [ Field.InArray "smallGroups" Table.User groupIds ]

            for user in users do
                do! Patch.byId Table.User user.Id {| SmallGroups = user.SmallGroups |> List.except groupIds |}

            do! Delete.byFields Table.Group All [ Field.Equal "churchId" churchId ]
            do! Delete.byId Table.Church churchId
            do! txn.CommitAsync()
        }

    /// Save a church's information
    let save church = save<Church> Table.Church church

    /// Find a church by its ID
    let tryById churchId =
        Find.byId<ChurchId, Church> Table.Church churchId


/// Functions to manipulate small group members
module Members =

    /// Count members for the given small group
    let countByGroup (groupId: SmallGroupId) =
        Count.byFields Table.Member All [ Field.Equal "smallGroupId" groupId ]

    /// Delete a small group member by its ID
    let deleteById (memberId: MemberId) = Delete.byId Table.Member memberId

    /// Retrieve all members for a given small group
    let forGroup (groupId: SmallGroupId) =
        Find.byFieldsOrdered<Member>
            Table.Member
            All
            [ Field.Equal "smallGroupId" groupId ]
            [ Field.Named "memberName" ]

    /// Save a small group member
    let save mbr = save<Member> Table.Member mbr

    /// Retrieve a small group member by its ID
    let tryById memberId =
        Find.byId<MemberId, Member> Table.Member memberId


/// Options to retrieve a list of requests
type PrayerRequestOptions =
    {
        /// The small group for which requests should be retrieved
        SmallGroup: SmallGroup

        /// The clock instance to use for date/time manipulation
        Clock: IClock

        /// The date for which the list is being retrieved
        ListDate: LocalDate option

        /// Whether only active requests should be retrieved
        ActiveOnly: bool

        /// The page number, for paged lists
        PageNumber: int
    }


/// Functions to manipulate prayer requests
module PrayerRequests =

    /// Central place to append sort criteria for prayer request queries
    let private orderBy sort =
        match sort with
        | SortByDate -> [ Field.Named "updatedDate DESC"; Field.Named "enteredDate DESC"; Field.Named "requestor" ]
        | SortByRequestor -> [ Field.Named "requestor"; Field.Named "updatedDate DESC"; Field.Named "enteredDate DESC" ]
        |> fun fields -> Query.orderBy fields SQLite

    /// Paginate a prayer request query
    let private paginate (pageNbr: int) pageSize =
        if pageNbr > 0 then
            $"LIMIT {pageSize} OFFSET {(pageNbr - 1) * pageSize}"
        else
            ""

    /// Count the number of prayer requests for a church
    let countByChurch churchId =
        backgroundTask {
            let! groupIds = SmallGroups.groupIdsByChurch churchId
            return! Count.byFields Table.Request All [ Field.In "smallGroupId" groupIds ]
        }

    /// Count the number of prayer requests for a small group
    let countByGroup (groupId: SmallGroupId) =
        Count.byFields Table.Request All [ Field.Equal "smallGroupId" groupId ]

    /// Delete a prayer request by its ID
    let deleteById (reqId: PrayerRequestId) = Delete.byId Table.Request reqId

    /// Get all (or active) requests for a small group as of now or the specified date
    let forGroup (opts: PrayerRequestOptions) =
        let theDate = defaultArg opts.ListDate (opts.SmallGroup.LocalDateNow opts.Clock)

        let sql, parameters =
            if opts.ActiveOnly then
                let expDate =
                    (theDate.AtStartOfDayInZone(opts.SmallGroup.TimeZone)
                         - Duration.FromDays opts.SmallGroup.Preferences.DaysToExpire)
                        .ToInstant()
                $"""AND (   data->>'updatedDate'  > :updatedDate
                         OR data->>'expiration'   = :expManual
                         OR data->>'requestType' IN (:typLongTerm, :typExpecting))
                    AND data->>'expiration' <> :expForced""",
                [ SqliteParameter(":updatedDate", expDate)
                  SqliteParameter(":expManual", string Manual)
                  SqliteParameter(":typLongTerm", string LongTermRequest)
                  SqliteParameter(":typExpecting", string Expecting)
                  SqliteParameter(":expForced", string Forced) ]
            else
                "", []

        Custom.list
            $"SELECT data FROM {Table.Request}
               WHERE data->>'smallGroupId = :groupId {sql}
               ORDER BY {orderBy opts.SmallGroup.Preferences.RequestSort}
               {paginate opts.PageNumber opts.SmallGroup.Preferences.PageSize}"
            (SqliteParameter(":groupId", string opts.SmallGroup.Id) :: parameters)
            fromData<PrayerRequest>

    /// Save a prayer request
    let save req = save<PrayerRequest> Table.Request req

    /// Search prayer requests for the given term
    let searchForGroup group searchTerm pageNbr =
        Custom.list
            $"SELECT data FROM {Table.Request}
               WHERE data->>'smallGroupId' = :groupId
                 AND data->>'requestText' LIKE :search
                  UNION
              SELECT data FROM {Table.Request}
               WHERE data->>'smallGroupId' = :groupId
                 AND COALESCE(data->>'requestor', '') LIKE :search
              ORDER BY {orderBy group.Preferences.RequestSort}
              {paginate pageNbr group.Preferences.PageSize}"
            [ SqliteParameter(":groupId", string group.Id)
              SqliteParameter(":search", $"%%%s{searchTerm}%%") ]
            fromData<PrayerRequest>

    /// Retrieve a prayer request by its ID
    let tryById reqId =
        Find.byId<PrayerRequestId, PrayerRequest> Table.Request reqId

    /// Update the expiration for the given prayer request
    let updateExpiration (req: PrayerRequest) withTime =
        if withTime then
            Patch.byId
                Table.Request
                req.Id
                {| UpdatedDate = req.UpdatedDate
                   Expiration = req.Expiration |}
        else
            Patch.byId Table.Request req.Id {| Expiration = req.Expiration |}


/// Functions to manipulate users
module Users =

    /// Retrieve all PrayerTracker users
    let all () =
        Find.allOrdered<User> Table.User [ Field.Named "lastName"; Field.Named "firstName" ]

    /// Count the number of users for a church
    let countByChurch churchId =
        backgroundTask {
            let! groupIds = SmallGroups.groupIdsByChurch churchId
            return! Count.byFields Table.User All [ Field.InArray "smallGroups" Table.User groupIds ]
        }

    /// Count the number of users for a small group
    let countByGroup (groupId: SmallGroupId) =
        Count.byFields Table.User All [ Field.InArray "smallGroups" Table.User [ groupId ] ]

    /// Delete a user by its database ID
    let deleteById (userId: UserId) = Delete.byId Table.User userId

    /// Get a list of users authorized to administer the given small group
    let listByGroupId (groupId: SmallGroupId) =
        Find.byFieldsOrdered<User>
            Table.User
            All
            [ Field.InArray "smallGroups" Table.User [ groupId ] ]
            [ Field.Named "lastName"; Field.Named "firstName" ]

    /// Save a user's information
    let save user = save<User> Table.User user

    /// Find a user by its e-mail address and authorized small group
    let tryByEmailAndGroup (email: string) (groupId: SmallGroupId) =
        Find.firstByFields<User>
            Table.User
            All
            [ Field.Equal "email" email
              Field.InArray "smallGroups" Table.User [ groupId ] ]

    /// Find a user by their database ID
    let tryById userId =
        Find.byId<UserId, User> Table.User userId

    /// Update a user's last seen date/time
    let updateLastSeen (userId: UserId) (now: Instant) =
        Patch.byId Table.User userId {| LastSeen = now |}

    /// Update a user's password hash
    let updatePassword (user: User) =
        Patch.byId Table.User user.Id {| PasswordHash = user.PasswordHash |}

    /// Update a user's authorized small groups
    let updateSmallGroups (userId: UserId) (groupIds: SmallGroupId list) =
        Patch.byId Table.User userId {| SmallGroups = groupIds |}
